(define (strict-array? obj)
  (and (array? obj) (not (string? obj)) (not (vector? obj))))

;;; matrix:sorta-partial-pivot
;;;
;;; Performs something vaguely recognizable as partial pivoting;
;;; it's not "real" partial pivoting, but it's an algorithm that
;;; is said to help.  Returns the permutation matrix consed onto
;;; the new matrix.

(define (matrix:sorta-partial-pivot m)
  (let* ((a (matrix:copy m))
	(rows (matrix:rows a))
	 (cols (matrix:cols a))
	(P (make-vector rows))
	 (magic-d
	  (lambda (i j)
	    (let* ((row (vector->list (matrix:row a i))))
	      (/ (abs (matrix:ref a i j))
		 (apply max (map abs row))))))
	 (pivot
	  (lambda ()
	    (do ((level 0 (1+ level))) ((= level (- rows 1)))
	      (let loop ((row (1+ level))
			 (max-row level)
			 (max-d (magic-d level level)))
		(cond
		 ((> row (- (matrix:rows a) 1))
		  (if (not (= max-row level))
		      (let ((temp-row (matrix:row a max-row))
			    (temp-n (vector-ref P max-row)))
			(matrix:set-row! a max-row (matrix:row a level))
			(matrix:set-row! a level temp-row)
			(vector-set! P max-row level)
			(vector-set! P level temp-n))))
		 ((> (magic-d row level) max-d)
		  (loop (1+ row) row (matrix:ref a row level)))
		 (else
		  (loop (1+ row) max-row max-d))))))))
    
    ;; Create initial permutation vector
    (do ((i 0 (1+ i))) ((= i rows))
      (vector-set! P i i))
    ;; Actually perform the pivoting
    (pivot)
    ;; Create the return value
    (cons P a)))

;;; matrix:plu-factor
;;;
;;; Returns the PLU factorization of the matrix.
;;; The permutation vector is consed onto the results
;;; of consing L onto U.
;;;
;;; Gaussian elimination with partial pivoting (rows only)
;;; is used to performs the factorization.  The results of
;;; switching rows are kept in a permutation vector which
;;; is returned as P in the PLU factorization.

(define (matrix:plu-factor m)
  (cond
   ((not (matrix? m))
    (matrix:error 'matrix:plu-factor "~s is not a matrix" m))
   ((not (= (matrix:rows m) (matrix:cols m)))
    (matrix:error 'matrix:plu-factor
		  "cannot factor a non-square matrix"))
   (else
    (let* ((PM (matrix:sorta-partial-pivot m))
	   (a (cdr PM))
	   (P (car PM))
		(rows (matrix:rows a))
		(cols (matrix:cols a)))
      
      ;; Main loop.  Goes top to bottom and left to right,
      ;; creating zeros as it goes.
      
      (do ((k 0 (1+ k))) ((= k (- rows 1)))
	(do ((i (1+ k) (1+ i))) ((= i rows))
	  (let ((m_ik (/ (matrix:ref a i k) (matrix:ref a k k))))
	    (matrix:set! a i k m_ik)
	    (do ((j (1+ k) (1+ j))) ((= j cols))
	      (matrix:set! a i j
			   (- (matrix:ref a i j)
			      (* m_ik (matrix:ref a k j))))))))
      
      ;; Create the return value.  This returns separate L and U
      ;; matrices by pulling the appropriate values out of a.
      (cons P (cons (matrix:create rows cols
				   (lambda (i j)
				     (cond
				      ((= i j) 1)
				      ((> i j) (matrix:ref a i j))
				      (else 0))))
		    (matrix:create rows cols
				   (lambda (i j)
				     (if (<= i j)
					 (matrix:ref a i j) 0)))))))))

;;; matrix?
;;; matrix??
;;;
;;; Return #t if the object is a matrix; #f otherwise.
;;;
;;; matrix? does not test very thoroughly.  This is
;;; because the full comparison is a little bit too slow.
;;; All this one does is make sure that the object is a
;;; pair whose car is 'matrix and whose cdr is a vector.
;;;
;;; matrix?? is the thorough one.  First, we make sure
;;; that the outer part is a vector, then make sure that
;;; the first element of that vector is again a vector.
;;; Once both of those are true, make sure that every other
;;; element is a vector with the same length as the the
;;; first one.

(define (matrix? a)
;;  (and (pair? a)
;;       (eq? (car a) 'matrix)
;;       (vector? (cdr a))))

	(strict-array? a))

;; define (matrix?? a)
;;  (and (matrix? a)
;;       (vector? (vector-ref (cdr a) 0))
;;       (let ((l (vector-length (vector-ref (cdr a) 0))))
;;  (= l (apply + (map (lambda (x)
;; 		      (and (vector? x)
;; 			   (if (= (vector-length x) l) 1 0)))
;; 		    (vector->list (cdr a))))))))

;;; matrix:rows
;;;
;;; Returns the number of rows in the matrix.

(define (matrix:rows a)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:rows "~s is not a matrix" a))
   (else
	(car (array-dimensions a)))))

;;; matrix:cols
;;;
;;; Returns the number of columns in the matrix.

(define (matrix:cols a)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:cols "~s is not a matrix" a))
   (else
	(car (cdr (array-dimensions a))))))

;;; matrix:copy
;;;
;;; Returns an identical copy of the given matrix.

(define (matrix:copy a) 
	(matrix:create (matrix:rows a) (matrix:cols a) 
		(lambda (i j)
			(matrix:ref a i j))))

;;; matrix:row
;;;
;;; Returns the ith row of matrix a as a vector.
;;;
;;; Note: The return vector is 0-based!

(define (matrix:row a i)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:row "~s is not a matrix" a))
   ((or (< i 0) (> i (matrix:rows a)))
    (matrix:error 'matrix:row
 	  "row subscript is out of range for array ~s" a))
   (else
	(list->vector (list-ref (array->list a) i)))))

;;; matrix:col
;;;
;;; Returns the jth column of matrix a as a vector.
;;;
;;; Note: The vector is 0-based!

;; define (matrix:col a j)
;;  (cond
;;   ((not (matrix? a))
;;    (matrix:error 'matrix:col "~s is not a matrix" a))
;;   ((or (< j 1) (> j (matrix:cols a)))
;;    (matrix:error 'matrix:col
;; 	  "column subscript is out of range for array ~s" a))
;;   (else
;;    (list->vector (map (lambda (v) (vector-ref v (1- j)))
;; 	       (vector->list (cdr a)))))))

;;; matrix:ref
;;;
;;; Returns the element a_ij.

(define (matrix:ref a i j)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:ref "~s is not a matrix" a))
   ((or (< i 0) (> i (matrix:rows a)))
    (matrix:error 'matrix:ref
 	  "row subscript ~s is out of range for matrix ~s" i a))
   ((or (< j 0) (> j (matrix:cols a)))
    (matrix:error 'matrix:ref
 	  "column subscript ~s is out of range for matrix ~s" j a))
   (else
	(array-ref a i j))))


;;; matrix:set!
;;;
;;; Set!'s the a_ij element of the matrix to value.

(define (matrix:set! a i j value)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:set! "~s is not a matrix" a))
   ((or (< i 0) (> i (matrix:rows a)))
    (matrix:error 'matrix:ref
		  "row subscript ~s is out of range for array ~s" i a))
   ((or (< j 0) (> j (matrix:cols a)))
    (matrix:error 'matrix:ref
		  "column subscript ~s is out of range for array ~s" j a))
   (else
	(array-set! a value i j))))

;;; matrix:create
;;;
;;; Creates a new matrix with the given dimensions.
;;; The element for each position is the result of
;;; the application of fn to the row and column
;;; values.  If fn is not a procedure, the value of
;;; fn is placed directly into the matrix.

(define (matrix:create row-size col-size fn)
  (cond
   ((or (not (integer? row-size)) (< row-size 1))
    (matrix:error 'matrix:create "row-size must be a positive integer"))
   ((or (not (integer? col-size)) (< col-size 1))
    (matrix:error 'matrix:create "col-size must be a positive integer"))
   (else
;;    (do ((a (make-vector row-size))
;;	 (r (1- row-size) (1- r)))
;;	((= r -1) (cons 'matrix a))
;;     (vector-set! a r (do ((v (make-vector col-size))
;;			    (c (1- col-size) (1- c)))
;;			   ((= c -1) v)
;;			 (vector-set! v c (if (procedure? fn)
;;					      (fn (1+ r) (1+ c))
;;					      fn))))))))
	(do ((a (make-array '#(0) row-size col-size))
		(r (1- row-size) (1- r)))
		((= r -1) a)
		(do ((c (1- col-size) (1- c))) 
			((= -1 c))(array-set! a (if (procedure? fn)
							(fn r c)
							fn) r c)))))) 

;; matrix:mset
;; 
;; Updates the content of matrix 
(define (matrix:mset! a fn)
	(let rloop ((i 0))
		(cond ((< i (matrix:rows a))
			(let cloop ((j 0))
				(cond ((< j (matrix:cols a))
					(matrix:set! a i j (if (procedure? fn)
						(fn i j) fn))
					(cloop (1+ j)))))
			(rloop (1+ i))))))

;;; matrix:set-row!
;;;
;;; Sets the given row of the given matrix
;;; to the given value.

(define (matrix:set-row! a i value)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:set-row! "~s is not a matrix" a))
   ((or (< i 0) (> i (matrix:rows a)))
    (matrix:error 'matrix:set-row! "row subscript ~s is out of range" i))
   ((not (= (matrix:cols a) (vector-length value)))
    (matrix:error 'matrix:set-row! "vector ~s is not of length ~s"
	   value (matrix:cols a)))
   (else
		(let loop ((j 0))
			(cond ((< j (matrix:cols a))
				(matrix:set! a i j (vector-ref value j))
				(loop (1+ j))))))))

;;; matrix:permute-rows
;;;
;;; Permutes the rows of matrix according to
;;; the permutation vector given.

(define (matrix:permute-rows a perm-vec)
  (cond
   ((not (matrix? a))
    (matrix:error 'matrix:permute-rows "~s is not a matrix" a))
   ((not (vector? perm-vec))
    (matrix:error 'matrix:permute-rows "~s is not a vector" perm-vec))
   ((not (= (vector-length perm-vec) (matrix:rows a)))
    (matrix:error 'matrix:permute-rows
	   "permutation vector ~s is the wrong length" perm-vec))
   (else
    (let ((new-a (matrix:copy a)))
      (do ((i 0 (1+ i))) ((= i (matrix:rows a)))
	(matrix:set-row! new-a i (matrix:row a (vector-ref perm-vec i))))
      new-a))))
